;==============================================================================
;
;    LONGNAME.ASM -- 16-bit lookalikes for selected WIN32 APIs
;
;    Written by Walter Oney
;
;    Functions:
;	 DosDateTimeToFileTime
;	 FindClose
;	 FindFirstFile
;	 FindNextFile
;	 FileTimeToSystemTime
;	 GetCurrentDirectory
;	 GetLastError
;	 GetVolumeInformationEx
;
;    Dependencies:
;	 386 or better processor (uses 32-bit registers, etc.)
;
;    Assemble with MASM 6.1:
;	 ml -c longname.asm
;
;==============================================================================

	 name  longname
	 .model small, c
	 .386

byp	 equ   <byte ptr>
wp	 equ   <word ptr>
dwp	 equ   <dword ptr>

;    WIN32_FIND_DATA structure:

fd		   struct
dwFileAttributes   dd ?
ftCreationTime	   dq ?
ftLastAccessTime   dq ?
ftLastWriteTime    dq ?
nFileSizeHigh	   dd ?
nFileSizeLow	   dd ?
dwReserved0	   dd ?
dwReserved1	   dd ?
cFileName	   db 260 dup (?)
cAlternateFileName db 14 dup (?)
fd		   ends

;    Disk information structure filled in by 21/69:

dinfo	 struc
infolevel dw   ?
volser	 dd    ?
volname  db    11 dup (?)
fsname	 db    8 dup (?)
nullchar db    ?		  ; padding (byte not used by 21/69)
dinfo	 ends

	 .data
lasterr  dw    0

	 .code

;------------------------------------------------------------------------------
;
;    HANDLE WINAPI FindFirstFile(LPTSTR lpszPath, LPWIN32_FIND_DATA lpFindData)
;
;------------------------------------------------------------------------------

FindFirstFile proc far pascal uses di, lpszPath:dword, lpFindData:dword
	 local hfind:word

	 mov   lasterr, 0	  ; assume call will succeed

	 push  ds		  ; save DS across call (1)
	 mov   ax, 714Eh	  ; fcn 714E: find first file (long name version)
	 lds   dx, lpszPath	  ; DS:DX -> search pattern
	 les   di, lpFindData	  ; ES:DI -> results buffer
	 mov   cx, 10h		  ; CX = attribute mask: files or directories

	 stc			  ; set carry in case fcn not supported
	 int   21h		  ; issue DOS request
	 jnc   goback		  ; skip ahead if success
	 cmp   ax, 7100h	  ; is function supported?
	 jne   failure		  ; if yes, we've got an error

;    For non-Chicago version of the API, allocate memory to hold the search
;    state and install it as the DTA. Then issue the 8.3 version of FindFirst.
;    The memory block handle becomes the find-file handle used by this
;    API

	 mov   bx, 3		  ; we need roughly 48 bytes of memory
	 mov   ah, 48h		  ; fcn 48: allocate memory
	 int   21h		  ; allocate memory to record search info
	 jc    failure		  ; die if error

	 mov   hfind, ax	  ; save buffer address in safe register
	 mov   ah, 2Fh		  ; fcn 2F:get DTA address
	 int   21h		  ; return in ES:BX
	 push  es		  ; save current DTA address on stack (2)
	 push  bx		  ;   ..

	 mov   ah, 1Ah		  ; fcn 1A: set DTA
	 mov   ds, hfind	  ; DS:DX -> new DTA
	 xor   dx, dx		  ;   ..
	 int   21h		  ;   ..

	 mov   ah, 4Eh		  ; fcn 4E: find first file
	 lds   dx, lpszPath	  ; DS:DX -> search pattern
	 mov   cx, 10h		  ; CX = attribute mask: files or directories
	 int   21h		  ; issue request
	 jc    failure_old	  ; skip ahead if error

	 pop   dx		  ; restore DTA (2)
	 pop   ds		  ;   ..
	 mov   ah, 1Ah		  ;   ..
	 int   21h		  ;   ..
	 pop   ds		  ; restore DS (1)

	 push  hfind		  ; copy info to caller's buffer
	 push  lpFindData	  ;   ..
	 call  findcopy 	  ;   ..

	 mov   ax, hfind	  ; return with find-file handle in AX
	 ret			  ; return to caller (DS already popped)

;    Return find-file handle to caller

goback:
	 pop   ds		  ; restore DS (1)
	 ret			  ; return with find-file handle in AX

;    Here on failure

failure:
	 mov   lasterr, ax	  ; save error code
	 xor   ax, ax		  ; indicate failure by returning null handle
	 jmp   goback		  ;   ..

failure_old:
	 mov   lasterr, ax	  ; save error code
	 pop   dx		  ; restore DTA (2)
	 pop   ds		  ;   ..
	 mov   ah, 1Ah		  ;   ..
	 int   21h		  ;   ..

	 mov   ah, 49h		  ; fcn 49: release memory
	 mov   es, hfind	  ; ES = memory block to free
	 int   21h		  ; release unused handle

	 xor   ax, ax		  ; indicate failure by returning null handle
	 jmp   goback		  ;   ..

FindFirstFile endp

;------------------------------------------------------------------------------
;
;    BOOL WINAPI FindNextFile(HANDLE hFindFile, LPWIN32_FIND_DATA lpFindData)
;
;------------------------------------------------------------------------------

FindNextFile proc far pascal uses di, hFindFile:word, lpFindData:dword
	 mov   lasterr, 0	  ; assume call will succeed

	 push  ds		  ; save DS across call (1)
	 mov   ax, 714Fh	  ; fcn 714F: find next file (long name version)
	 mov   bx, hFindFile	  ; BX = find-file handle
	 les   di, lpFindData	  ; ES:DI -> results buffer

	 stc			  ; set carry in case fcn not supported
	 int   21h		  ; issue DOS request
	 jnc   success		  ; skip ahead if success
	 cmp   ax, 7100h	  ; is function supported?
	 jne   failure		  ; if yes, we've got an error

;    Use find-file handle as the DTA during the 8.3 function call:

	 mov   ah, 2Fh		  ; fcn 2F:get DTA address
	 int   21h		  ; return in ES:BX
	 push  es		  ; save current DTA address on stack (2)
	 push  bx		  ;   ..

	 mov   ah, 1Ah		  ; fcn 1A: set DTA
	 mov   ds, hFindFile	  ; DS:DX -> new DTA
	 xor   dx, dx		  ;   ..
	 int   21h		  ;   ..

	 mov   ah, 4Fh		  ; fcn 4E: find next file
	 int   21h		  ; issue request
	 jc    failure_old	  ; skip ahead if error

	 pop   dx		  ; restore DTA (2)
	 pop   ds		  ;   ..
	 mov   ah, 1Ah		  ;   ..
	 int   21h		  ;   ..
	 pop   ds		  ; restore DS (1)

	 push  hFindFile	  ; push args to findcopy
	 push  lpFindData	  ;   ..
	 call  findcopy 	  ; copy data to caller's buffer

	 mov   ax, 1		  ; return TRUE (DS already popped)
	 ret

;    Here to return to our caller

success:
	 mov   ax, 1		  ; return TRUE to indicate one more match

goback:
	 pop   ds		  ; restore DS (1)
	 ret			  ; return to caller

failure:
	 mov   lasterr, ax	  ; save last error
	 xor   ax, ax		  ; return FALSE to indicate no more matches
	 jmp   goback		  ; done

failure_old:
	 mov   lasterr, ax	  ; save last error
	 pop   dx		  ; restore DTA (2)
	 pop   ds		  ;   ..
	 mov   ah, 1Ah		  ;   ..
	 int   21h		  ;   ..
	 xor   ax, ax		  ; return FALSE to indicate no more matches
	 jmp   goback		  ;   ..

FindNextFile endp

;------------------------------------------------------------------------------
;
;    FINDCOPY: Copy find-file data from 8.3 buffer to WIN32 buffer
;
;------------------------------------------------------------------------------

findcopy proc  pascal private uses di si, hfind:word, lpFindData:dword

	 mov   es, hfind	  ; push date & time
	 push  wp es:[18h]	  ;   ..
	 push  wp es:[16h]	  ;   ..
	 les   di, lpFindData	  ; push address of last write time
	 lea   di, [di].fd.ftLastWriteTime
	 push  es		  ;   ..
	 push  di		  ;   ..
	 call  DosDateTimeToFileTime ; convert last-write timestamp

	 push  ds		  ; save DS (1)

	 mov   ds, hfind	  ; DS:0 -> find file data
	 les   di, lpFindData	  ; ES:DI -> WIN32 find-file buffer

	 movzx eax, byp ds:[15h]  ; copy attribute
	 mov   es:[di].fd.dwFileAttributes, eax

	 mov   eax, dwp ds:[1Ah]  ; copy file size
	 mov   es:[di].fd.nFileSizeLow, eax
	 mov   es:[di].fd.nFileSizeHigh, 0

	 mov   es:[di].fd.cAlternateFileName, 0 ; flags cFileName as 8.3

	 mov   dwp es:[di].fd.ftCreationTime, 0
	 mov   dwp es:[di].fd.ftCreationTime+4, 0
	 mov   dwp es:[di].fd.ftLastAccessTime, 0
	 mov   dwp es:[di].fd.ftLastAccessTime+4, 0
	 mov   dwp es:[di].fd.dwReserved0, 0
	 mov   dwp es:[di].fd.dwReserved1, 0

	 lea   di, [di].fd.cFileName ; es:di -> where to put filename
	 mov   si, 1Eh		     ; ds:si -> filename we found
	 mov   cx, 13		     ; cx = length (13 bytes)
	 rep   movsb		     ; copy filename

	 pop   ds		  ; restore DS (1)
	 ret			  ; return to local caller

findcopy endp

;------------------------------------------------------------------------------
;
;    BOOL WINAPI FindClose(HANDLE hFindFile)
;
;------------------------------------------------------------------------------

FindClose proc far pascal, hFindFile:word
	 mov   lasterr, 0	  ; assume call will succeed

	 mov   ax, 71A1h	  ; fcn 71A1: find close (long name version)
	 mov   bx, hFindFile	  ; BX = find-file handle

	 stc			  ; set carry in case fcn not supported
	 int   21h		  ; issue DOS request
	 jnc   success		  ; skip ahead if success
	 cmp   ax, 7100h	  ; is function supported?
	 jne   failure		  ; if yes, we've got an error

	 mov   ah, 49h		  ; fcn 49: release memory block
	 mov   es, bx		  ; BX = memory block handle
	 int   21h		  ; go release memory

success:
	 mov   ax, 1		  ; return TRUE to indicate success

goback:
	 ret			  ; return to caller

failure:
	 mov   lasterr, ax	  ; save last error
	 xor   ax, ax		  ; return FALSE to indicate failure
	 jmp   goback		  ; done

FindClose endp

;------------------------------------------------------------------------------
;
;    BOOL WINAPI GetCurrentDirectory(DWORD cbPath, LPTSTR lpszPath)
;
;------------------------------------------------------------------------------

GetCurrentDirectory proc far pascal uses si di, cbPath:dword, lpszPath:dword
	 local dirname[260]:byte

	 mov   lasterr, 0	  ; assume call will succeed

	 mov   ah, 19h		  ; fcn 19: get default disk drive
	 int   21h		  ; issue DOS request
	 add   al, 'A'            ; convert 0 to A, etc.
	 mov   dirname, al	  ; build initial part of name
	 mov   dirname+1, ':'     ;   ..
	 mov   dirname+2, '\'     ;   ..

	 push  ds		  ; save DS across call
	 mov   ax, 7147h	  ; fcn 7147: get current directory (long name version)
	 xor   dl, dl		  ; DL = drive (0 means default drive)
	 mov   si, ss		  ; DS:SI -> buffer to receive name
	 mov   ds, si		  ;   ..
	 lea   si, dirname+3	  ;   ..

	 stc			  ; set carry in case not supported
	 int   21h		  ; issue DOS request
	 jnc   success		  ; skip if call succeed in long-name form
	 cmp   ax, 7100h	  ; check for bona fide error
	 jne   failure		  ; skip if so

	 mov   ah, 47h		  ; reissue request in 8.3 form
	 int   21h		  ;   ..
	 jc    failure		  ; skip if error

success:
	 les   di, lpszPath	  ; ES:DI -> destination buffer
	 mov   cx, wp cbPath	  ; CX = length of destination incl null
	 cld			  ; should already be clear, but make sure
	 lea   si, dirname	  ; DS:SI -> start of current directory

@@:
	 test  cx, cx		  ; exhausted buffer count yet?
	 jz    toosmall 	  ; if yes, leave loop
	 lodsb			  ; copy a byte from our buffer to destination
	 stosb			  ;   ..
	 dec   cx		  ; reduce residual buffer count
	 test  al, al		  ; reached null terminator yet?
	 jnz   @B		  ; if not, continue

;    Indicate success by returning the number of bytes copied to the caller's
;    buffer. If we're returning the root directory, the length is zero. The
;    caller can tell this isn't an error by calling GetLastError. Otherwise,
;    the return value will be nonzero but less than the cbPath argument
;    because we don't include the trailing null terminator.

	 mov   ax, wp cbPath	  ; get buffer length again
	 sub   ax, cx		  ; # bytes we copied
	 dec   ax		  ; (don't include null terminator)
	 pop   ds		  ; restore DS

goback:
	 ret			  ; return to caller

;    Indicate failure by returning zero. Caller can get the error code by
;    calling GetLastError.

failure:
	 pop   ds		  ; restore DS
	 mov   lasterr, ax	  ; save error code
	 xor   ax, ax		  ; set return code zero to indicate failure
	 jmp   goback		  ; done

;    Here when the caller's buffer is too small to hold the string. We've
;    already copied cbPath bytes. DS:SI points to the first byte we weren't
;    able to copy.

toosmall:
	 pop   ds		  ; restore DS
	 mov   di, ss		  ; set ES:DI -> uncopied string
	 mov   es, di		  ;   ..
	 mov   di, si		  ;   ..
	 xor   al, al		  ; get zero to compare against
	 dec   cx		  ; get infinite count
	 repne scasb		  ; scan for null terminator

	 neg   cx		  ; get number of characters left in string
	 dec   cx		  ;   ..
	 dec   cx		  ;   ..
	 add   cx, wp cbPath	  ; include the stuff we were able to copy
	 mov   ax, cx		  ; set AX = return value
	 jmp   goback		  ; return to caller

GetCurrentDirectory endp

;------------------------------------------------------------------------------
;
;    DWORD GetLastError()
;
;------------------------------------------------------------------------------

GetLastError proc far pascal
	 mov   ax, lasterr	  ; get last error code
	 ret			  ; return to caller
GetLastError endp

;------------------------------------------------------------------------------
;
;    BOOL WINAPI GetVolumeInformationEx(LPTSTR lpRoot, LPTSTR lpVolName,
;	DWORD cbVolName, LPDWORD lpVolSer, LPDWORD lpMaxName, LPDWORD lpMaxPath,
;	LPDWORD lpFlag, LPTSTR lpFSName, DWORD cbFSName);
;
;    Note: The "Ex" part of this function names comes from the lpMaxPath
;    argument I added to the calling sequence of the WIN32 function. Watch
;    out for Microsoft adding a new API under this name with a different
;    calling sequence!
;
;------------------------------------------------------------------------------

GetVolumeInformationEx proc far pascal uses di si, lpRoot:dword, \
			    lpVolName:dword, cbVolName:dword, \
			    lpVolSer:dword, lpMaxName:dword, \
			    lpMaxPath:dword, lpFlag:dword, \
			    lpFSName:dword, cbFSName:dword

	 local diskinfo:dinfo	  ; to fill in via 21/69

	 mov   lasterr, 0	  ; assume function will succeed
	 push  ds		  ; save DS across DOS calls (1)

;    Set DS:DX to the name of the root directory for the volume we're
;    interested in.

	 lds   dx, lpRoot	  ; DS:DX -> name of root directory
	 mov   ax, ds		  ; is pointer null?
	 or    ax, dx		  ;   ..
	 jnz   @F		  ; if not, good

	 mov   dx, cs		  ; yes. use "\" instead
	 mov   ds, dx		  ;   ..
	 mov   dx, offset root	  ;   ..
@@:

;    Use undocumented function 21/69 to get volume info and the file
;    system name. This function isn't translated by any of the DOS
;    extenders (including Windows) I tested in "Testing the Windows DOS
;    Extender", Windows/DOS Developer's Journal, Vol. 5, No. 2 (Feb. 1994),
;    pp. 47-60, but nothing catastrophic happened either

	 push  ds		  ; save root name across call (2)
	 push  dx		  ;   ..
	 mov   bx, dx		  ; BL = drive number
	 cmp   byp [bx+1], ':'    ; colon in string?
	 mov   bl, byp [bx]	  ; get 1st character in any case
	 je    @F		  ; skip if there's a colon
	 mov   bl, 'a' - 1        ; no colon, so ask for default drive
@@:
	 or    bl, ' '            ; convert drive letter to lower case
	 sub   bl, 'a' - 1        ; BL = drive number (0 = default, 1 = A)

	 mov   dx, ss		  ; DS:DX -> results buffer
	 mov   ds, dx		  ;   ..
	 lea   dx, diskinfo	  ;   ..

	 mov   ax, 6900h	  ; fcn 6900: get disk serial number
	 int   21h		  ; issue DOS request
	 pop   dx		  ; restore root name in any case (2)
	 pop   ds		  ;   ..
	 jc    failure		  ; die if error in call

;    First try the Chicago GetVolumeInformation call. Note that the NT API
;    allows a NULL lpRoot parameter, whereas the INT 21 function does not.

	 les   di, lpFSName	  ; ES:DI -> buffer to receive file system name
	 mov   cx, wp cbFSName	  ; CX = length of that buffer
	 mov   ax, 71A0h	  ; fcn 71A0: get volume information
	 stc			  ; set carry in case fcn not supported

	 int   21h		  ; issue DOS request
	 jnc   setflags 	  ; skip ahead if no error
	 cmp   ax, 7100h	  ; is function supported?
	 jne   failure		  ; if not, error of some kind

;    21/71A0 isn't supported by the operating system. Store values that
;    indicate we're dealing with a vanilla FAT file system, but get the
;    file system name from 21/69

	 mov   si, ss		  ; set DS:SI -> file system name from 21/69
	 mov   ds, si		  ;   ..
	 lea   si, diskinfo.fsname;   ..
	 mov   diskinfo.nullchar, 0 ; add a null terminator

	 call  strncpy		  ; copy file system name to caller's buffer
	 xor   bx, bx		  ; flags = 0
	 mov   cx, 13		  ; length of xxxxxxxx.xxx0
	 mov   dx, 81		  ; maximum length of a pathname

;    Store the flags and capacities

setflags:
	 les   di, lpMaxName	  ; save maximum name component length
	 movzx ecx, cx		  ;   ..
	 mov   dwp es:[di], ecx   ;   ..
	 les   di, lpMaxPath	  ; save maximum path length
	 movzx edx, dx		  ;   ..
	 mov   dwp es:[di], edx   ;   ..
	 les   di, lpFlag	  ; save file system capability flags
	 movzx ebx, bx		  ;   ..
	 mov   dwp es:[di], ebx   ;   ..

;    If requested, fill in volume name and volume serial number from the
;    previous results of 21/69

	 les   di, lpVolName	  ; ES:DI -> volume name buffer
	 mov   ax, es		  ; is pointer null?
	 or    ax, di		  ;   ..
	 jz    @F		  ; if yes, don't copy anything

	 mov   cx, wp cbVolName   ; CX = length of buffer
	 mov   si, ss		  ; DS:SI -> volume name from 21/69
	 mov   ds, si		  ;   ..
	 lea   si, diskinfo.volname ; ..
	 mov   diskinfo.fsname, 0 ; force a null terminator
	 call  strncpy		  ; copy volume name to caller's buffer
@@:
	 les   di, lpVolSer	  ; ES:DI -> volume serial number
	 mov   ax, es		  ; is pointer NULL?
	 or    ax, di		  ;   ..
	 jz    @F		  ; if yes, don't copy anything
	 mov   eax, diskinfo.volser ; copy volume serial number
	 mov   dwp es:[di], eax   ;   ..
@@:
	 jmp   success		  ; done -- everything worked

;    Here when one of the calls fails

failure:
	 pop   ds		  ; restore DS (1)
	 mov   lasterr, ax	  ; save error code
	 xor   ax, ax		  ; return FALSE to indicate failure
	 jmp   goback		  ;   ..

success:
	 pop   ds		  ; restore DS (1)
	 mov   ax, 1		  ; return TRUE to indicate success

goback:
	 ret			  ; return to caller

root	 db    '\', 0             ; default root directory name

GetVolumeInformationEx endp

strncpy  proc  private
@@:
	 test  cx, cx		  ; is buffer count down to zero?
	 jz    @F		  ; if yes, leave loop
	 lodsb			  ; no. copy one character
	 stosb			  ;   ..
	 dec   cx		  ; reduce buffer count
	 test  al, al		  ; copied null terminator yet?
	 jnz   @B		  ; if not, continue
@@:
	 ret
strncpy  endp

;==================== D A T E	C O N V E R S I O N S =========================

;    Table of first days of each month (with Jan 1 being zero) for leap years:

	 .data

monthday dw    0
	 dw    0+31
	 dw    0+31+29
	 dw    0+31+29+31
	 dw    0+31+29+31+30
	 dw    0+31+29+31+30+31
	 dw    0+31+29+31+30+31+30
	 dw    0+31+29+31+30+31+30+31
	 dw    0+31+29+31+30+31+30+31+31
	 dw    0+31+29+31+30+31+30+31+31+30
	 dw    0+31+29+31+30+31+30+31+31+30+31
	 dw    0+31+29+31+30+31+30+31+31+30+31+30
	 dw    0+31+29+31+30+31+30+31+31+30+31+30+31

FILETIME struct
dwLowDateTime  dd ?
dwHighDateTime dd ?
FILETIME ends

SYSTEMTIME struct
wYear	       dw ?
wMonth	       dw ?
wDayOfWeek     dw ?
wDay	       dw ?
wHour	       dw ?
wMinute        dw ?
wSecond        dw ?
wMilliseconds  dw ?
SYSTEMTIME ends

	 .code

;------------------------------------------------------------------------------
;
;    BOOL WINAPI DosDateTimeToFileTime(WORD wFatDate, WORD wFatTime,
;	LPFILETIME lpFileTime);
;
;------------------------------------------------------------------------------

DosDateTimeToFileTime proc far pascal uses di si, wFatDate:word, wFatTime:word,\
			       lpFileTime:dword

	 local dayindex:dword, ftime:dword

	 mov   dx, wFatDate	    ; get date
	 mov   ebx, edx 	    ; isolate month index 1-12 (bits 5-8)
	 and   ebx, 01E0h	    ;	..
	 shr   ebx, 4		    ;	..
	 movzx ecx, [monthday-2+bx] ; get base day as if Feb had 29 days

	 mov   eax, edx 	    ; isolate day number 1-n (bits 0-4)
	 and   eax, 001Fh	    ;	..
	 lea   ecx, [eax+ecx-1]     ; ECX => day within this year

;  Accumulate leap-year adjustment flag in EBX. We started the calculation
;  by acting as though every year had a Feb 29, so that March 1 was the
;  61'st day of the year. We'll subtract one day at the end of the
;  algorithm if the day is >= March 1 and it's not a leap year.
;
;  The bytes of EBX are used as follows:
;
;     0 -- true if not last year of a quad-century (which is a leap year)
;     1 -- true if last year of a century (not usually a leap year)
;     2 -- true if year not divisible by four (can't be a leap year)
;     3 -- true if day on or after March 1
;
;  The adjustment value is "1" iff: 3 and (2 or (1 and 0))

	 cmp   ecx, 31+29	    ; is it after Feb. 29 of this year?
	 setge bh		    ; if yes, set BH = 1 (adjust in non-leap year)

	 mov   eax, edx 	    ; isolate year (bits 9-15, offset from 1980)
	 and   eax, 0FE00h	    ;	..
	 shr   eax, 9		    ;	..
	 add   eax, 1980-1601	    ; EAX = year index since time origin

	 xor   edx, edx 	    ; calculate year index within quad-year
	 mov   edi, 4		    ;	(i.e., 4 years per quad-year)
	 div   edi		    ; EAX = # quad-years since origin, EDX = rem
	 cmp   edx, 3		    ; is this the last year of a quad-year?
	 setne bl		    ; if yes, set BL = 0 (leap year unless last of non-leap century)
	 shl   ebx, 16		    ; save intermediate terms in high half of EBX

	 push  eax		    ; save # quad-years since origin (1a)
	 mov   eax, edx 	    ; calculate # days since start of quad year
	 mov   edi, 365 	    ;	(i.e., 365 days per non-leap year)
	 mul   edi		    ;	..
	 add   ecx, eax 	    ; ECX => # days since start of quad-year

	 pop   eax		    ; restore # quad-years since origin (1a)
	 xor   edx, edx 	    ; calculate quad-year index within century
	 mov   edi, 25		    ;	(i.e., 25 quad-years per century)
	 div   edi		    ; EAX = # centuries since origin, EDX = rem
	 cmp   edx, 24		    ; is this the last quad-year of a century?
	 sete  bl		    ; if yes, set BL = 1 (only a leap year if this is leap century)

	 push  eax		    ; save # centuries since origin (1b)
	 mov   eax, edx 	    ; calculate # days since start of century
	 mov   edi, 4*365+1	    ;	(i.e., # days in a quad-year ending in a leap year)
	 mul   edi		    ;	..
	 add   ecx, eax 	    ; ECX => # days since start of century

	 pop   eax		    ; restore # centuries since origin (1b)
	 xor   edx, edx 	    ; compute century index within quad-century
	 mov   edi, 4		    ;	(i.e., 4 centuries per quad-century)
	 div   edi		    ; EAX = # quad-centuries since origin (who cares??)
	 cmp   edx, 3		    ; is this final century of quad-century?
	 setne bh		    ; if yes, set BH = 0 (last year is a leap year)

	 push  eax		    ; save # quad-centuries since origin (1c)
	 mov   eax, edx 	    ; compute # days since start of quad-century
	 mov   edi, 25*(4*365+1)-1  ;	(i.e., # days in century ending in non-leap year)
	 mul   edi		    ;	..
	 add   ecx, eax 	    ; ECX => # days since start of quad-century

	 pop   eax		    ; restore # quad-centuries since origin (1c)
	 mov   edi, 4*(25*(4*365+1)-1)+1 ; (i.e., # days in quad-century ending in leap year)
	 mul   edi		    ;	..
	 add   ecx, eax 	    ; ECX => # days since Jan. 1, 1601 time origin

	 and   bh, bl		    ; calculate leap year adjustment: 0 and 1
	 shr   ebx, 8		    ;	..
	 or    bh, bl		    ;	... or 2
	 shr   ebx, 8		    ;	..
	 and   bl, bh		    ;	... and 3
	 movzx ebx, bl		    ; EBX now is 1 or 0
	 sub   ecx, ebx 	    ; adjust day number to account for short Februaries

      ;  At this point, ECX holds the number of days since the time origin at
      ;  Jan 1, 1601.

	 mov   bx, wFatTime	    ; EBX = packed file time
	 mov   dayindex, ecx	    ; save day count
	 mov   ecx, ebx 	    ; isolate seconds in ECX (bits 0-4)
	 and   ecx, 001Fh	    ;	..
	 add   ecx, ecx 	    ;	..

	 mov   eax, ebx 	    ; isolate minutes (bits 5-10)
	 and   eax, 07E0h	    ;	..
	 shr   eax, 5		    ;	..
	 mov   edi, 60		    ; calculate # seconds since start of hour
	 mul   edi		    ;	..
	 add   ecx, eax 	    ; accumulate seconds in ECX

	 mov   eax, ebx 	    ; isolate hours (bits 11-15)
	 and   eax, 0F800h	    ;	..
	 shr   eax, 11		    ;	..
	 mov   edi, 60*60	    ; calculate # seconds since start of hour
	 mul   edi		    ;	..

	 add   eax, ecx 	    ; get # 4-second units in EAX, remainder in EDX
	 xor   edx, edx 	    ;	..
	 mov   edi, 4		    ;	..
	 div   edi		    ; EAX = # 4-second units today, EDX = # seconds after that
	 mov   ftime, eax	    ; save # 4-second units

	 mov   eax, edx 	    ; calculate # 100-nanoseconds in given 4-second span
	 mov   edi, 10000000	    ;	(i.e., # 100-nanoseconds in one second)
	 mul   edi		    ;	..
	 mov   ecx, eax 	    ; save in ECX for later accumulation

      ;  Now we have ECX = # 100-nanoseconds in particular 4-second span corresponding
      ;  to this time value.

	 mov   eax, dayindex	    ; get day index
	 mov   edi, 6*60*60	    ;	(i.e., # 4-second units in one day)
	 mul   edi		    ; EAX = # 4-second units from origin to start of today
	 add   eax, ftime	    ; now # 4-second units from origin to file's time
	 mov   edi, 40000000	    ; calculate # 100-nanoseconds since origin
	 mul   edi		    ;	..
	 add   eax, ecx 	    ; add in the last little bit
	 adc   edx, 0		    ;  ..

	 les   di, lpFileTime
	 mov   es:[di].FILETIME.dwLowDateTime, eax
	 mov   es:[di].FILETIME.dwHighDateTime, edx

	 mov   ax, 1		    ; return TRUE
	 ret			    ;	..

DosDateTimeToFileTime endp

;------------------------------------------------------------------------------
;
;    BOOL WINAPI FileTimeToSystemTime(LPFILETIME lpft, LPSYSTEMTIME lpst);
;
;------------------------------------------------------------------------------

FileTimeToSystemTime proc far pascal uses di si, lpft:dword, lpst:dword

	 les   di, lpft 	    ; ES:DI -> file time (ns since Jan 1, 1601)
	 lgs   si, lpst 	    ; GS:SI -> system time assembly buffer
	 mov   eax, es:[di].FILETIME.dwLowDateTime ; EDX:EAX = file time
	 mov   edx, es:[di].FILETIME.dwHighDateTime

	 mov   edi, 40000000	    ; EAX => time in units of 4 sec, EDX = remainder (100-ns)
	 div   edi
	 mov   ecx, eax 	    ; save quotient in EAX

	 mov   eax, edx 	    ; compute milliseconds portion of time
	 xor   edx, edx 	    ;	..
	 mov   edi, 10000	    ;	(i.e., milliseconds per 100-nanosecond)
	 div   edi		    ;	..
	 mov   edi, 1000	    ;	..
	 xor   edx, edx 	    ;	..
	 div   edi		    ; EAX = seconds, EDX = rem (ms)
	 mov   gs:[si].SYSTEMTIME.wMilliseconds, dx
	 xchg  ecx, eax 	    ; save seconds in ECX, recover time in 4-sec units since origin

	 xor   edx, edx 	    ; calculate minutes since origin
	 mov   edi, 15		    ;	(i.e., 15 units of 4 seconds per minute)
	 div   edi		    ;	..
	 shl   edx, 2		    ; convert remainder to seconds
	 add   edx, ecx 	    ; total number of seconds
	 mov   gs:[si].SYSTEMTIME.wSecond, dx ; will be < 60

	 xor   edx, edx 	    ; calculate hours since origin
	 mov   edi, 60		    ;	..
	 div   edi		    ;	(i.e., 60 minutes per hour)
	 mov   gs:[si].SYSTEMTIME.wMinute, dx

	 xor   edx, edx 	    ; calculate days since origin
	 mov   edi, 24
	 div   edi		    ;	(i.e., 24 hours per day)
	 mov   gs:[si].SYSTEMTIME.wHour, dx

	 mov   ecx, eax 	    ; save days since origin
	 add   eax, 1		    ; 1/1/1601 was a Monday
	 xor   edx, edx 	    ; so, what is today?
	 mov   edi, 7
	 div   edi	       ;   ..
	 mov   gs:[si].SYSTEMTIME.wDayOfWeek, dx

	 mov   eax, ecx 	    ; recover days since origin
	 xor   edx, edx 	    ; compute quad-century since origin
	 mov   edi, 4*(25*(4*365+1)-1)+1 ; (i.e., # days in quad-centry ending in leap year)
	 div   edi
	 mov   ah, 100
	 mul   ah	   ; convert to 0, 400, etc., multiple
	 shl   eax, 2		    ;	..
	 lea   ecx, [eax+1601]	    ; ECX = base year of quad-century

	 mov   eax, edx 	    ; recover # days since start of quad-century
	 xor   edx, edx 	    ; compute century index within quad-century
	 mov   edi, 25*(4*365+1)-1	 ;   (i.e., # days in century not ending in leap year
	 div   edi
	 cmp   eax, 3		    ; is it the last century?
	 setne bl		    ; BL = 0 if last century (ends in leap year)

	 mov   ah, 100		    ; convert to 0, 100, etc., multiple
	 mul   ah
	 add   ecx, eax 	    ; ECX => base year of century
	 mov   eax, edx 	    ; recover # days since start of century
	 xor   edx, edx 	    ; compute quad-year index within century
	 mov   edi, 4*365+1		 ;   (i.e., # days in quad-year ending in leap-year
	 div   edi

	 cmp   eax, 24		    ; is this the last quad-year of the century?
	 sete  bh		    ; if yes, BH = 1 (not a leap year unless it's 2000, 2400, etc.)
	 and   bl, bh		    ; BL = 0 if this is a leap quad-year

	 mov   ah, 4		    ; convert to 0, 4, etc., multiple
	 mul   ah
	 add   ecx, eax 	    ; EAX => base year of quad-year
	 mov   eax, edx 	    ; recover # days since start of quad-year
	 xor   edx, edx 	    ; calculate year index within quad-year
	 mov   edi, 365
	 div   edi		    ;	(i.e, # days in non-leap year)

	 cmp   eax, 3		    ; is this the last year of the quad year?
	 setne bh		    ; if not, BH = 1
	 or    bl, bh		    ; BL = 0 if this is a leap year

	 add   ecx, eax 	    ; ECX => final year value
	 mov   gs:[si].SYSTEMTIME.wYear, cx

	 mov   eax, edx 	    ; recover day index within year
	 cmp   eax, 31+28	    ; is it after Feb. 28?
	 setge bh		    ; if yes, BH = 1
	 and   bl, bh		    ; BL = 1 if > Feb 28 and not a leap year
	 movzx ebx, bl		    ; adjust day as if all Februarys had 29 days
	 add   eax, ebx 	    ;	..

	 mov   bx, 2		    ; set month index = 1

@@:
	 cmp   ax, monthday[bx]     ; is day before start of next month?
	 jl    @F		    ; if yes, leave loop
	 add   bx, 2		    ; no. up to next month
	 jmp   @B

@@:
	 mov   cx, bx		    ; compute month index
	 shr   cx, 1		    ;	..
	 mov   gs:[si].SYSTEMTIME.wMonth, cx
	 inc   ax		    ; get day index 1-x
	 sub   ax, [monthday-2+bx]
	 mov   gs:[si].SYSTEMTIME.wDay, ax

	 mov   ax, 1		    ; return TRUE
	 ret			    ;	..

FileTimeToSystemTime endp

	 end
