;==============================================================================
;
;    REPLACE.ASM -- Visual C++ long filename replacement functions
;
;    Written by Walter Oney
;
;    Functions:
;	_chdir
;	_dos_creat
;	_dos_creatnew
;	_dos_getfileattr
;	_dos_open
;	_dos_setfileattr
;	_mkdir
;	_open
;	remove (_unlink)
;	rename
;	_rmdir
;	_sopen
;
;    Dependencies:
;	 386 or better processor (because of near jmps)
;	 DOS 4 or later (uses 21/6C)
;
;    Assemble with MASM 6.1:
;	 ml -c -DMODEL={SMALL|LARGE|MEDIUM|COMPACT|TINY} replace.asm
;
;==============================================================================

	 name  replace
	 .model MODEL, c
	 .386

byp	 equ   <byte ptr>
wp	 equ   <word ptr>
dwp	 equ   <dword ptr>

	 .data
	 extrn _osmajor:byte
	 extrn _fmode:word
	 extrn _umaskval:word
	 extrn _nfile:word
	 extrn _osfile:byte

	 .code

	 if @CodeSize
	   extrn _dosreturn:far
	   extrn _dosret0:far
	   extrn _dosretax:far
	 else
	   extrn _dosreturn:far
	   extrn _dosret0:near
	   extrn _dosretax:near
	 endif

;------------------------------------------------------------------------------
;
;    int _chdir(char *dirname);
;    int _mkdir(char *dirname);
;
;------------------------------------------------------------------------------

_mkdir	 proc  dirname:ptr byte
	 mov   ax, 7139h	  ; fcn 7139: make directory (long name)
	 jmp   dirop		  ; join common code with _chdir
_mkdir	 endp

_chdir	 proc  dirname:ptr byte
	 mov   ax, 713Bh	  ; fcn 713B: change directory (long name)

dirop::
	 if @DataSize
	   push ds		  ; save DS across call
	   lds dx, dirname	  ; DS:DX -> directory name
	 else
	   mov dx, dirname	  ;   ..
	 endif

	 mov   bl, al		  ; save function code
	 stc			  ; set carry in case fcn not supported
	 int   21h		  ; issue long filename form of request
	 jnc   goback		  ; skip ahead if success
	 cmp   ax, 7100h	  ; did it fail because fcn not supported?
	 jne   failure		  ; if not, error

	 mov   ah, bl		  ; reissue 8.3 form of request
	 int   21h		  ;   ..

goback:
	 jmp   _dosret0 	  ; return 0 or error code

failure:
	 stc			  ; force carry on
	 jmp   goback		  ; return with error code in AX
_chdir	 endp

;------------------------------------------------------------------------------
;
;    unsigned _dos_creat(char *filename, unsigned attrib, int *handle);
;
;------------------------------------------------------------------------------

_dos_creat proc filename:ptr byte, attrib:word, handle:ptr word
	 push  si		  ; save working registers
	 push  di		  ;   ..

	 mov   di, 3C00h	  ; use 21/3C if long name doesn't work
	 mov   dx, 12h		  ; DL: replace if exists, create if doesn't
	 mov   bx, 2		  ; BH (flags) = 0, BL = R/W open in compatability mode

create::
	 if @DataSize
	   push ds		  ; save DS across calls
	 endif

	 mov   ax, 716Ch	  ; fcn 716C: extended create (long name)
	 mov   cx, attrib	  ; CX = attributes for new file

	 if @DataSize
	   lds si, filename	  ; DS:SI -> name of file to create
	 else
	   mov si, filename	  ;   ..
	 endif

	 stc			  ; set carry in case not supported
	 int   21h		  ; issue long filename request
	 jnc   success		  ; skip ahead if success
	 cmp   ax, 7100h	  ; is long filename call supported?
	 jne   failure		  ; if not, we've got an error

	 mov   ax, di		  ; reissue request as 21/3C or 21/5B
	 mov   dx, si		  ; DS:DX -> name of file
	 int   21h		  ;   ..
	 jc    failure		  ; skip on error

success:
	 if @DataSize
	   lds si, handle	  ; set DS:SI -> handle buffer
	 else
	   mov si, handle	  ;   ..
	 endif

	 mov   wp [si], ax	  ; save file handle

goback:
	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 pop   di		  ; restore working regs
	 pop   si		  ;   ..
	 jmp   _dosreturn	  ; return with 0 or error code in AX

failure:
	 cmp   al, 50h		  ; is the error "file exists" from 21/5B?
	 jne   @F		  ; if not, okay
	 mov   ah, 11h		  ; yes. substitute internal error code
@@:
	 stc			  ; set carry to flag error
	 jmp   goback		  ; return with error code in AX

_dos_creat endp

;------------------------------------------------------------------------------
;
;    unsigned _dos_createnew(char *filename, unsigned attrib, int *handle);
;
;------------------------------------------------------------------------------

_dos_creatnew proc filename:ptr byte, attrib:word, handle:ptr word
	 push  si		  ; save working registers
	 push  di		  ;   ..

	 mov   di, 5B00h	  ; use 21/5B if long name fails
	 mov   dx, 10h		  ; DL: fail if exists, create if doesn't
	 mov   bx, 2		  ; BH (flags) = 0, BL = R/W open in compatability mode
	 jmp   create		  ; join common code above
_dos_creatnew endp

;------------------------------------------------------------------------------
;
;    unsigned _dos_getfileattr(char *filename, unsigned *attrib);
;
;------------------------------------------------------------------------------

_dos_getfileattr proc filename:ptr byte, attrib:ptr byte

	 if @DataSize
	   push ds		  ; save DS across calls
	 endif

	 mov   ax, 7143h	  ; fcn 7143: get/set attributes (long name)
	 xor   bl, bl		  ; BL = 0 to indicate "get" fcn

	 if @DataSize
	   lds dx, filename	  ; DS:DX -> name of file to investigate
	 else
	   mov dx, filename	  ;   ..
	 endif

	 stc			  ; set carry in case not supported
	 int   21h		  ; issue long filename form of request
	 jnc   success		  ; skip ahead on success
	 cmp   ax, 7100h	  ; is fcn supported?
	 jne   failure		  ; if not, we've got error

	 mov   ax, 4300h	  ; fcn 4300: get attribute
	 int   21h		  ; issue 8.3 form of request
	 jc    goback		  ; skip ahead if error

success:
	 if @DataSize
	   lds bx, attrib	  ; get attribute byte address
	 else
	   mov bx, attrib	  ;   ..
	 endif

	 mov   wp [bx], cx	  ; store attribute byte from CX
	 xor   ax, ax		  ; indicate successful call

goback:
	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 jmp   _dosreturn	  ; return with 0 or error code

failure:
	 stc			  ; be sure carry flag is set
	 jmp   goback		  ; return with error code in AX

_dos_getfileattr endp

;------------------------------------------------------------------------------
;
;    unsigned _dos_open(char *filename, unsigned mode, int *handle);
;
;------------------------------------------------------------------------------

_dos_open proc filename:ptr byte, mode:word, handle:ptr word
	 push  si		  ; save working registers
	 push  di		  ;   ..

	 mov   bx, mode 	  ; fetch open mode into BL
	 mov   bh, 3Dh		  ; use 21/3Dxx if long-name fails
	 mov   di, bx		  ;   ..
	 xor   bh, bh		  ; clear flags in BH
	 mov   dx, 01h		  ; DL: open if exists, fail if doesn't
	 jmp   create		  ; join common code above
_dos_open endp

;------------------------------------------------------------------------------
;
;    unsigned _dos_setfileattr(char *filename, unsigned attrib);
;
;------------------------------------------------------------------------------

_dos_setfileattr proc filename:ptr byte, attrib:word

	 if @DataSize
	   push ds		  ; save DS across calls
	 endif

	 mov   ax, 7143h	  ; fcn 7143: get/set attributes (long name)
	 mov   bl, 1		  ; BL = 1 to indicate "set" fcn
	 mov   cx, attrib	  ; CX = new attributes for file

	 if @DataSize
	   lds dx, filename	  ; DS:DX -> name of file to change
	 else
	   mov dx, filename	  ;   ..
	 endif

	 stc			  ; set carry in case not supported
	 int   21h		  ; issue long filename form of request
	 jnc   goback		  ; skip ahead on success
	 cmp   ax, 7100h	  ; is fcn supported?
	 jne   failure		  ; if not, we've got error

	 mov   ax, 4301h	  ; fcn 4301: set attribute
	 int   21h		  ; issue 8.3 form of request

goback:
	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 jmp   _dosreturn	  ; return with 0 or error code in AX

failure:
	 stc			  ; be sure carry is set
	 jmp   goback		  ; return with error code in AX

_dos_setfileattr endp

;------------------------------------------------------------------------------
;
;    int remove(char *filename);
;    int _unlink(char *filename);
;
;------------------------------------------------------------------------------

	 public remove
remove:

_unlink  proc  filename:ptr byte

	 if @DataSize
	   push ds		  ; save DS across calls
	   lds	dx, filename	  ; DS:DX -> name of file to delete
	 else
	   mov	dx, filename	  ;   ..
	 endif

	 mov   ax, 7141h	  ; fcn 7141: delete file
	 stc			  ; set carry in case not supported
	 int   21h		  ; issue long filename form of request
	 jnc   goback		  ; skip ahead if okay
	 cmp   ax, 7100h	  ; is fcn supported?
	 jne   failure		  ; if not, error

	 mov   ah, 41h		  ; issue 8.3 form of request
	 int   21h		  ;   ..

goback:
	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 jmp   _dosret0 	  ; return 0 or error code

failure:
	 stc			  ; be sure CF is set
	 jmp   goback		  ; done

_unlink  endp

;------------------------------------------------------------------------------
;
;    int rename(char *oldname, char *newname);
;
;------------------------------------------------------------------------------

rename	 proc  oldname:ptr byte, newname:ptr byte

	 push  di		  ; save working register

	 if @DataSize
	   push ds		  ; save DS across calls
	   lds dx, oldname	  ; DS:DX -> old name
	   les di, newname	  ; ES:DI -> new name
	 else
	   mov dx, oldname	  ;   ..
	   mov di, ds		  ;   ..
	   mov es, di		  ;   ..
	   mov di, newname	  ;   ..
	 endif

	 mov   ax, 7156h	  ; fcn 7156: long name form of rename
	 stc			  ; set carry in case not supported
	 int   21h		  ; issue long filename form of rename
	 jnc   done		  ; skip ahead on success
	 cmp   ax, 7100h	  ; is fcn supported?
	 jne   failure		  ; if not, error

	 mov   ah, 56h		  ; resissue 8.3 form of request
	 int   21h		  ;   ..

done:
	 pop   di		  ; restore working register

	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 jmp   _dosret0 	  ; return with zero or error code

failure:
	 stc			  ; set carry again to indicate error
	 jmp   done

rename	 endp

;------------------------------------------------------------------------------
;
;    int _rmdir(char *dirname);
;
;------------------------------------------------------------------------------

_rmdir	 proc  dirname:ptr byte

	 if @DataSize
	   push ds		  ; save DS
	   lds	dx, dirname	  ; DS:DX -> directory name
	 else
	   mov	dx, dirname	  ;   ..
	 endif

	 mov   ax, 713Ah	  ; fcn 713A: remove directory (long name)
	 stc			  ; set carry in case not supported
	 int   21h		  ; issue long name version of request
	 jnc   goback		  ; skip if no error
	 cmp   ax, 7100h	  ; did it fail because it's not supported?
	 jne   failure		  ; if not, handle error

	 mov   ah, 3Ah		  ; reissue 8.3 form of request
	 int   21h		  ;   ..
	 jc    failure		  ; skip to handle any error

goback:
	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 jmp   _dosret0 	  ; return 0 or error code

;    Analyze failure and revise error code. If we got 10 (attempt to remove
;    current directory), it might be because there are wild cards in the
;    name instead of because it's really the current directory. (Somebody
;    must have reported this as a bug to Microsoft once upon a time and
;    they decided to fix it.)

failure:
	 cmp   ax, 10h		  ; did we try to delete current directory?
	 jne   failure_2	  ; if not, okay
	 xchg  ax, dx		  ; save error code in DX
	 xchg  ax, bx		  ; DS:BX -> name of directory we tried to delete
@@:
	 mov   al, [bx] 	  ; fetch next character from dir name
	 inc   bx		  ;   ..
	 test  al, al		  ; reached null terminator?
	 jz    failure_1	  ; if yes, don't edit error code
	 cmp   al, '?'            ; is it a wild-card character?
	 je    @F		  ; if yes, change to path-not-found
	 cmp   al, '*'            ;   ..
	 jne   @B		  ;   ..
@@:
	 mov   dx, 3		  ; change to "path not found"

failure_1:
	 mov   ax, dx		  ; bring error code back to AX

failure_2:
	 stc			  ; force carry on
	 jmp   goback		  ; return with error code

_rmdir	 endp

;------------------------------------------------------------------------------
;
;    int _open(char *filename, int oflag [, int pmode]);
;    int _sopen(char *filename, int oflag, int shflag [, int pmode]);
;
;------------------------------------------------------------------------------

;    Open flags (from fcntl.h):

_O_RDONLY      equ 0000h
_O_WRONLY      equ 0001h
_O_RDWR        equ 0002h
_O_APPEND      equ 0008h

_O_CREAT       equ 0100h
_O_TRUNC       equ 0200h
_O_EXCL        equ 0400h

_O_TEXT        equ 4000h
_O_BINARY      equ 8000h

;    Sharing flags (from share.h):

_SH_COMPAT     equ 00h
_SH_DENYRW     equ 10h
_SH_DENYWR     equ 20h
_SH_DENYRD     equ 30h
_SH_DENYNO     equ 40h

;    Permission flags (from stat.h):

_S_IWRITE      equ 0080h
_S_IREAD       equ 0100h

_sopen	 proc  uses si, filename:ptr byte, oflag:word, shflag:word, pmode:word
	 local fileflag:byte, junk:byte
	 xor   bx, bx		  ; assume share flags irrelevant
	 cmp   _osmajor, 3	  ; DOS 3.0 or later?
	 jb    @F		  ; if not, that's right
	 mov   bl, byp shflag	  ; yes. pickup sharing arg
@@:
	 mov   ax, pmode	  ; copy optional permission setting
	 mov   shflag, ax	  ;   down to where _open can find it
	 jmp   open_1		  ; join common code with _open
_sopen	 endp

_open	 proc  uses si, filename:ptr byte, oflag:word, pmode:word
	 local fileflag:byte, junk:byte
	 xor   bx, bx		  ; share flags: compatibility mode

open_1::
	 mov   ax, oflag	  ; pickup open flags
	 mov   fileflag, 0	  ; assume binary access
	 mov   dx, 01h		  ; open if exists, fail if doesn't

;    Select binary or text. If none specified, use the global setting

	 test  ax, _O_BINARY	  ; is binary specified?
	 jnz   open_2		  ; if yes, okay
	 test  ax, _O_TEXT	  ; how about text?
	 jnz   @F		  ; if yes, also okay
	 test  _fmode, _O_BINARY  ; is default to use binary?
	 jnz   open_2		  ; if yes, skip ahead
@@:
	 mov   byp fileflag, 80h  ; set flag for text access

;    Compose open mode and action flags for eventual 21/716C

open_2:
	 and   al, 3		  ; preserve open mode 0, 1, or 2
	 or    bl, al		  ; merge with sharing flags left in BL

	 test  ax,_O_TRUNC	  ; truncate existing file?
	 jz    @F		  ; if not, skip
	 mov   dl, 02h		  ; yes. set action flag accordingly
@@:
	 test  ax, _O_CREAT	  ; might we create a missing file?
	 jz    open_3		  ; if not, skip ahead

	 or    dl, 10h		  ; yes. set action flag
	 test  ax, _O_EXCL	  ; fail open if file exists?
	 jz    @F		  ; if not, action is okay
	 and   dl, not 0Fh	  ; yes. modify action accordingly
@@:
	 mov   cx, pmode	  ; get permission mode argument
	 call  _cXENIXtoDOSmode   ; convert to DOS format

;    Setup and issue long filename extended open request. The
;    registers have already been setup as follows:
;	 BH (flags) = 0
;	 BL = open mode and sharing flags
;	 CX = attributes for new file
;	 DH (reserved) = 0
;	 DL = action to take if file exists or doesn't exist

open_3:
	 if @DataSize
	   push ds		  ; save DS across call
	   lds si, filename	  ; DS:SI -> name of file to open
	 else
	   mov si, filename	  ;   ..
	 endif

	 mov   ax, 716Ch	  ; fcn 716C: extended open (long filename)
	 stc			  ; set carry in case not supported
	 int   21h		  ; issue long filename open request

	 jnc   success		  ; skip ahead if it succeeds
	 cmp   ax, 7100h	  ; did it fail because fcn not supported?
	 jne   failure		  ; if yes, quit due to error

;    Long filename request isn't supported, so reissue the request using the
;    8.3 extended open function. This won't work on pre-DOS 4 systems (because
;    21/6C didn't exist then) or under certain DOS extenders. If this is a
;    problem for you, reverse engineer the "open" module in the runtime
;    library and use the requisite combination of simpler functions.

	 mov   ax, 6C00h	  ; reissue request in 8.3 form
	 int   21h		  ;   ..
	 jnc   success		  ; skip ahead if error

failure:
	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 cmp   al, 50h		  ; is error 50, meaning file already exists?
	 jne   failure_1	  ; if not, okay
	 mov   ax, 1100h	  ; yes. substitute internal error code

failure_1:
	 stc			  ; force carry flag on
	 pop   si		  ; restore working register
	 jmp   _dosretax	  ; record error & return -1

;    Here when open succeeds. AX is currently the file handle. Use 21/4400
;    to see if we're dealing with a device or a disk file.

success:
	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 mov   bx, ax		  ; save file handle in BX
	 mov   ax, 4400h	  ; fcn 4400: get device information
	 int   21h		  ;    ..
	 and   dl, 80h		  ; keep "device" flag only
	 shr   dl, 1		  ; shift to position
	 or    fileflag, dl	  ; merge with text-access flag

	 test  dl, dl		  ; is it a device?
	 jnz   setflag		  ; if yes, don't look for ^Z

;    If we're opening a disk file for text access, see if file ends
;    in a ^Z EOF marker. If so, truncate the file just before it.
;    We only do this for RDWR access, though I'm not sure why.

	 test  fileflag, 80h	  ; text access?
	 jz    setflag		  ; if not, no problem here
	 test  oflag, _O_RDWR	  ; read-write access?
	 jz    setflag		  ; if not, there won't be any confusion

	 mov   cx, -1		  ; CX:DX = file position (-1)
	 mov   dx, cx		  ;   ..
	 mov   ax, 4202h	  ; fcn 4202: seek relative to EOF
	 int   21h		  ; position file to end

	 neg   cx		  ; convert -1 to +1

	 if @DataSize
	   push ds		  ; save DS across call
	   mov	dx, ss		  ; set DS = SS
	   mov	ds, dx		  ;   ..
	 endif

	 lea   dx, junk 	  ; DS:DX -> 1-byte data buffer
	 mov   ah, 3Fh		  ; fcn 3F: read file
	 int   21h		  ; read last byte of file

	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 test  ax, ax		  ; test return code
	 jz    @F		  ; skip ahead if nothing read
	 cmp   junk, 1Ah	  ; or if byte isn't a ^Z
	 jne   @F		  ;   ..

	 neg   cx		  ; seek back to EOF-1
	 mov   dx, cx		  ;   ..
	 mov   ax, 4202h	  ;   ..
	 int   21h		  ;   ..

	 xor   cx, cx		  ; set data count = 0 to truncate file
	 mov   ah, 40h		  ; fcn 40: write file
	 int   21h		  ; issue request to truncate before ^Z
@@:
	 xor   cx, cx		  ; seek back to start of file
	 xor   dx, dx		  ;   ..
	 mov   ax, 4200h	  ;   ..
	 int   21h		  ;   ..

;    If we've opened a disk file, see if it's read-only

setflag:
	 test  fileflag, 40h	  ; is this a device?
	 jnz   setflag_1	  ; if yes, that's a good assumption

	 if @DataSize
	   push ds		  ; save DS across get-attr call
	   lds dx, filename	  ; DS:DX -> name of file
	 else
	   mov dx, filename	  ;   ..
	 endif

	 mov   ax, 7143h	  ; fcn 7143: get/set attribute
	 push  bx		  ; save file handle
	 xor   bl, bl		  ; BL = 0 to indicate "get"
	 stc			  ; set carry in case not supported
	 int   21h		  ; go get file attribute
	 pop   bx		  ; restore file handle in any case
	 jnc   @F		  ; skip ahead if no error

	 mov   ax, 4300h	  ; reissue 8.3 form of request
	 int   21h		  ;   ..
@@:
	 if @DataSize
	   pop ds		  ; restore DS
	 endif

	 test  cl, 1		  ; is file read-only?
	 jz    @F		  ; if yes, okay
	 or    fileflag, 10h	  ; yes. remember that
@@:
	 test  oflag, _O_APPEND   ; opening for append?
	 jz    setflag_1	  ; if not, okay
	 or    fileflag, 20h	  ; yes. remember that too

;    Save the appropriate internal access flag for this file:
;	 80    text-mode access
;	 40    file is really a device
;	 20    append to disk file
;	 10    read-only disk file
;	 01    file is open

setflag_1:
	 cmp   bx, _nfile	  ; is file handle within range of table?
	 jae   toomany		  ; if not, error
	 mov   cl, fileflag	  ; compose flag byte
	 or    cl, 1		  ; indicate file is open
	 mov   _osfile[bx], cl	  ; store flags

	 mov   ax, bx		  ; set file handle for normal return
	 ret			  ; return to external caller

toomany:
	 mov   ah, 3Eh		  ; fcn 3E: close file
	 int   21h		  ; issue close request
	 mov   ax, 1800h	  ; set internal error code
	 jmp   failure_1	  ; exit with error

_open	 endp

_cXENIXtoDOSmode proc
	 mov   ax, _umaskval	  ; get mask from _umask
	 not   ax		  ; invert because of negative sense of mask
	 and   ax, cx		  ; AX = bits we won't keep
	 xor   cx, cx		  ; clear those bits in CX
	 test  ax, _S_IWRITE	  ; is write permission allowed?
	 jnz   @F		  ; if yes, okay
	 or    cl, 1		  ; no. set read-only attribute
@@:
	 ret			  ; return to caller
_cXENIXtoDOSmode endp

	 end
